@@ -1,13 +1,17 @@
-/* dnsmasq is Copyright (c) 2000 - 2005 Simon Kelley
+/* dnsmasq is Copyright (c) 2000-2007 Simon Kelley
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 dated June, 1991.
-
+   the Free Software Foundation; version 2 dated June, 1991, or
+   (at your option) version 3 dated 29 June, 2007.
+ 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
+     
+  You should have received a copy of the GNU General Public License
+  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "dnsmasq.h"
@@ -21,7 +25,7 @@ static struct frec *lookup_frec_by_sender(unsigned short id,
 static unsigned short get_id(int force, unsigned short force_id, unsigned int crc);
 
 
-/* Send a UDP packet with it's source address set as "source" 
+/* Send a UDP packet with its source address set as "source" 
    unless nowild is true, when we just send it with the kernel default */
 static void send_from(int fd, int nowild, char *packet, size_t len, 
 		      union mysockaddr *to, struct all_addr *source,
@@ -107,7 +111,7 @@ static void send_from(int fd, int nowild, char *packet, size_t len,
     }
 }
           
-static unsigned short search_servers(struct daemon *daemon, time_t now, struct all_addr **addrpp, 
+static unsigned short search_servers(time_t now, struct all_addr **addrpp, 
 				     unsigned short qtype, char *qdomain, int *type, char **domain)
 			      
 {
@@ -140,7 +144,7 @@ static unsigned short search_servers(struct daemon *daemon, time_t now, struct a
 		  *addrpp = (struct all_addr *)&serv->addr.in6.sin6_addr;
 #endif 
 	      }
-	    else if (!flags)
+	    else if (!flags || (flags & F_NXDOMAIN))
 	      flags = F_NOERR;
 	  } 
       }
@@ -161,9 +165,9 @@ static unsigned short search_servers(struct daemon *daemon, time_t now, struct a
 	      flags = F_NXDOMAIN;
 	    else if (serv->flags & SERV_LITERAL_ADDRESS)
 	      {
-		if ((sflag | F_QUERY ) & qtype)
+		if (sflag & qtype)
 		  {
-		    flags = qtype & ~F_BIGNAME;
+		    flags = sflag;
 		    if (serv->addr.sa.sa_family == AF_INET) 
 		      *addrpp = (struct all_addr *)&serv->addr.in.sin_addr;
 #ifdef HAVE_IPV6
@@ -171,37 +175,36 @@ static unsigned short search_servers(struct daemon *daemon, time_t now, struct a
 		      *addrpp = (struct all_addr *)&serv->addr.in6.sin6_addr;
 #endif
 		  }
-		else if (!flags)
+		else if (!flags || (flags & F_NXDOMAIN))
 		  flags = F_NOERR;
 	      }
 	  } 
       }
 
-  if (flags & ~(F_NOERR | F_NXDOMAIN)) /* flags set here means a literal found */
-    {
-      if (flags & F_QUERY)
-	log_query(F_CONFIG | F_FORWARD | F_NEG, qdomain, NULL, 0, NULL, 0);
-      else
-	log_query(F_CONFIG | F_FORWARD | flags, qdomain, *addrpp, 0, NULL, 0);
-    }
-  else if (qtype && !(qtype & F_BIGNAME) && 
-	   (daemon->options & OPT_NODOTS_LOCAL) && !strchr(qdomain, '.') && namelen != 0)
-    /* don't forward simple names, make exception from NS queries and empty name. */
+  if (flags == 0 && !(qtype & F_BIGNAME) && 
+      (daemon->options & OPT_NODOTS_LOCAL) && !strchr(qdomain, '.') && namelen != 0)
+    /* don't forward simple names, make exception for NS queries and empty name. */
     flags = F_NXDOMAIN;
     
-  if (flags == F_NXDOMAIN && check_for_local_domain(qdomain, now, daemon))
+  if (flags == F_NXDOMAIN && check_for_local_domain(qdomain, now))
     flags = F_NOERR;
 
-  if (flags == F_NXDOMAIN || flags == F_NOERR)
-    log_query(F_CONFIG | F_FORWARD | F_NEG | qtype | (flags & F_NXDOMAIN), qdomain, NULL, 0, NULL, 0);
+  if (flags)
+    {
+      int logflags = 0;
+      
+      if (flags == F_NXDOMAIN || flags == F_NOERR)
+	logflags = F_NEG | qtype;
+  
+      log_query(logflags | flags | F_CONFIG | F_FORWARD, qdomain, *addrpp, 0, NULL, 0);
+    }
 
   return  flags;
 }
 
-/* returns new last_server */	
-static void forward_query(struct daemon *daemon, int udpfd, union mysockaddr *udpaddr,
-			  struct all_addr *dst_addr, unsigned int dst_iface,
-			  HEADER *header, size_t plen, time_t now, struct frec *forward)
+static int forward_query(int udpfd, union mysockaddr *udpaddr,
+			 struct all_addr *dst_addr, unsigned int dst_iface,
+			 HEADER *header, size_t plen, time_t now, struct frec *forward)
 {
   char *domain = NULL;
   int type = 0;
@@ -218,6 +221,7 @@ static void forward_query(struct daemon *daemon, int udpfd, union mysockaddr *ud
     {
       /* retry on existing query, send to all available servers  */
       domain = forward->sentto->domain;
+      forward->sentto->failed_queries++;
       if (!(daemon->options & OPT_ORDER))
 	{
 	  forward->forwardall = 1;
@@ -231,9 +235,9 @@ static void forward_query(struct daemon *daemon, int udpfd, union mysockaddr *ud
   else 
     {
       if (gotname)
-	flags = search_servers(daemon, now, &addrp, gotname, daemon->namebuff, &type, &domain);
+	flags = search_servers(now, &addrp, gotname, daemon->namebuff, &type, &domain);
       
-      if (!flags && !(forward = get_new_frec(daemon, now, NULL)))
+      if (!flags && !(forward = get_new_frec(now, NULL)))
 	/* table full - server failure. */
 	flags = F_NEG;
       
@@ -311,6 +315,7 @@ static void forward_query(struct daemon *daemon, int udpfd, union mysockaddr *ud
 			      (struct all_addr *)&start->addr.in6.sin6_addr, 0,
 			      NULL, 0);
 #endif 
+		  start->queries++;
 		  forwarded = 1;
 		  forward->sentto = start;
 		  if (!forward->forwardall) 
@@ -327,7 +332,7 @@ static void forward_query(struct daemon *daemon, int udpfd, union mysockaddr *ud
 	}
       
       if (forwarded)
-	  return;
+	return 1;
       
       /* could not send on, prepare to return */ 
       header->id = htons(forward->orig_id);
@@ -341,10 +346,10 @@ static void forward_query(struct daemon *daemon, int udpfd, union mysockaddr *ud
       send_from(udpfd, daemon->options & OPT_NOWILD, (char *)header, plen, udpaddr, dst_addr, dst_iface);
     }
 
-  return;
+  return 0;
 }
 
-static size_t process_reply(struct daemon *daemon, HEADER *header, time_t now, 
+static size_t process_reply(HEADER *header, time_t now, 
 			    struct server *server, size_t n)
 {
   unsigned char *pheader, *sizep;
@@ -389,7 +394,7 @@ static size_t process_reply(struct daemon *daemon, HEADER *header, time_t now,
     {
       if (header->rcode == NXDOMAIN && 
 	  extract_request(header, n, daemon->namebuff, NULL) &&
-	  check_for_local_domain(daemon->namebuff, now, daemon))
+	  check_for_local_domain(daemon->namebuff, now))
 	{
 	  /* if we forwarded a query for a locally known name (because it was for 
 	     an unknown type) and the answer is NXDOMAIN, convert that to NODATA,
@@ -399,7 +404,11 @@ static size_t process_reply(struct daemon *daemon, HEADER *header, time_t now,
 	  header->rcode = NOERROR;
 	}
       
-      extract_addresses(header, n, daemon->namebuff, now, daemon);
+      if (extract_addresses(header, n, daemon->namebuff, now))
+	{
+	  my_syslog(LOG_WARNING, _("possible DNS-rebind attack detected"));
+	  munged = 1;
+	}
     }
   
   /* do this after extract_addresses. Ensure NODATA reply and remove
@@ -419,7 +428,7 @@ static size_t process_reply(struct daemon *daemon, HEADER *header, time_t now,
 }
 
 /* sets new last_server */
-void reply_query(struct serverfd *sfd, struct daemon *daemon, time_t now)
+void reply_query(struct serverfd *sfd, time_t now)
 {
   /* packet from peer server, extract data for cache, and send to
      original requester */
@@ -467,7 +476,7 @@ void reply_query(struct serverfd *sfd, struct daemon *daemon, time_t now)
 		{
 		  header->qr = 0;
 		  header->tc = 0;
-		  forward_query(daemon, -1, NULL, NULL, 0, header, nn, now, forward);
+		  forward_query(-1, NULL, NULL, 0, header, nn, now, forward);
 		  return;
 		}
 	    }
@@ -489,7 +498,8 @@ void reply_query(struct serverfd *sfd, struct daemon *daemon, time_t now)
 		    break;
 		  }
 	    } 
-	  daemon->last_server = server;
+	  if (!(daemon->options & OPT_ALL_SERVERS))
+	    daemon->last_server = server;
 	}
       
       /* If the answer is an error, keep the forward record in place in case
@@ -499,7 +509,7 @@ void reply_query(struct serverfd *sfd, struct daemon *daemon, time_t now)
       if (forward->forwardall == 0 || --forward->forwardall == 1 || 
 	  (header->rcode != REFUSED && header->rcode != SERVFAIL))
 	{
-	  if ((nn = process_reply(daemon, header, now, server, (size_t)n)))
+	  if ((nn = process_reply(header, now, server, (size_t)n)))
 	    {
 	      header->id = htons(forward->orig_id);
 	      header->ra = 1; /* recursion if available */
@@ -511,7 +521,7 @@ void reply_query(struct serverfd *sfd, struct daemon *daemon, time_t now)
     }
 }
 
-void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
+void receive_query(struct listener *listen, time_t now)
 {
   HEADER *header = (HEADER *)daemon->packet;
   union mysockaddr source_addr;
@@ -531,6 +541,9 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
 #endif
 #if defined(HAVE_LINUX_NETWORK)
     char control[CMSG_SPACE(sizeof(struct in_pktinfo))];
+#elif defined(IP_RECVDSTADDR) && defined(HAVE_SOLARIS_NETWORK)
+    char control[CMSG_SPACE(sizeof(struct in_addr)) +
+		 CMSG_SPACE(sizeof(unsigned int))];
 #elif defined(IP_RECVDSTADDR)
     char control[CMSG_SPACE(sizeof(struct in_addr)) +
 		 CMSG_SPACE(sizeof(struct sockaddr_dl))];
@@ -598,7 +611,11 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
 	    if (cmptr->cmsg_level == IPPROTO_IP && cmptr->cmsg_type == IP_RECVDSTADDR)
 	      dst_addr_4 = dst_addr.addr.addr4 = *((struct in_addr *)CMSG_DATA(cmptr));
 	    else if (cmptr->cmsg_level == IPPROTO_IP && cmptr->cmsg_type == IP_RECVIF)
+#ifdef HAVE_SOLARIS_NETWORK
+	      if_index = *((unsigned int *)CMSG_DATA(cmptr));
+#else
 	      if_index = ((struct sockaddr_dl *)CMSG_DATA(cmptr))->sdl_index;
+#endif
 	}
 #endif
       
@@ -628,7 +645,7 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
 	return;
 #endif
       
-      if (!iface_check(daemon, listen->family, &dst_addr, &ifr, &if_index))
+      if (!iface_check(listen->family, &dst_addr, &ifr, &if_index))
 	return;
       
       if (listen->family == AF_INET &&
@@ -651,20 +668,26 @@ void receive_query(struct listener *listen, struct daemon *daemon, time_t now)
 #endif
     }
 
-  m = answer_request (header, ((char *) header) + PACKETSZ, (size_t)n, daemon, 
+  m = answer_request (header, ((char *) header) + PACKETSZ, (size_t)n, 
 		      dst_addr_4, netmask, now);
   if (m >= 1)
-    send_from(listen->fd, daemon->options & OPT_NOWILD, (char *)header, m, &source_addr, &dst_addr, if_index);
+    {
+      send_from(listen->fd, daemon->options & OPT_NOWILD, (char *)header, 
+		m, &source_addr, &dst_addr, if_index);
+      daemon->local_answer++;
+    }
+  else if (forward_query(listen->fd, &source_addr, &dst_addr, if_index,
+			 header, (size_t)n, now, NULL))
+    daemon->queries_forwarded++;
   else
-    forward_query(daemon, listen->fd, &source_addr, &dst_addr, if_index,
-		  header, (size_t)n, now, NULL);
+    daemon->local_answer++;
 }
 
 /* The daemon forks before calling this: it should deal with one connection,
    blocking as neccessary, and then return. Note, need to be a bit careful
    about resources for debug mode, when the fork is suppressed: that's
    done by the caller. */
-unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
+unsigned char *tcp_request(int confd, time_t now,
 			   struct in_addr local_addr, struct in_addr netmask)
 {
   int size = 0;
@@ -672,7 +695,7 @@ unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
   unsigned short qtype, gotname;
   unsigned char c1, c2;
   /* Max TCP packet + slop */
-  unsigned char *packet = malloc(65536 + MAXDNAME + RRFIXEDSZ);
+  unsigned char *packet = whine_malloc(65536 + MAXDNAME + RRFIXEDSZ);
   HEADER *header;
   struct server *last_server;
   
@@ -708,8 +731,11 @@ unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
 	}
       
       /* m > 0 if answered from cache */
-      m = answer_request(header, ((char *) header) + 65536, (unsigned int)size, daemon, 
+      m = answer_request(header, ((char *) header) + 65536, (unsigned int)size, 
 			 local_addr, netmask, now);
+
+      /* Do this by steam now we're not in the select() loop */
+      check_log_writer(NULL); 
       
       if (m == 0)
 	{
@@ -719,7 +745,7 @@ unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
 	  char *domain = NULL;
 	  
 	  if (gotname)
-	    flags = search_servers(daemon, now, &addrp, gotname, daemon->namebuff, &type, &domain);
+	    flags = search_servers(now, &addrp, gotname, daemon->namebuff, &type, &domain);
 	  
 	  if (type != 0  || (daemon->options & OPT_ORDER) || !daemon->last_server)
 	    last_server = daemon->servers;
@@ -754,7 +780,8 @@ unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
 		  
 		  if ((last_server->tcpfd == -1) &&
 		      (last_server->tcpfd = socket(last_server->addr.sa.sa_family, SOCK_STREAM, 0)) != -1 &&
-		      connect(last_server->tcpfd, &last_server->addr.sa, sa_len(&last_server->addr)) == -1)
+		      (!local_bind(last_server->tcpfd,  &last_server->source_addr, last_server->interface, 1) ||
+		       connect(last_server->tcpfd, &last_server->addr.sa, sa_len(&last_server->addr)) == -1))
 		    {
 		      close(last_server->tcpfd);
 		      last_server->tcpfd = -1;
@@ -762,7 +789,7 @@ unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
 		  
 		  if (last_server->tcpfd == -1)	
 		    continue;
-		  
+
 		  c1 = size >> 8;
 		  c2 = size;
 		  
@@ -776,7 +803,7 @@ unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
 		      last_server->tcpfd = -1;
 		      continue;
 		    } 
-	      
+		  
 		  m = (c1 << 8) | c2;
 		  if (!read_write(last_server->tcpfd, packet, m, 1))
 		    return packet;
@@ -799,7 +826,7 @@ unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
 		     someone might be attempting to insert bogus values into the cache by 
 		     sending replies containing questions and bogus answers. */
 		  if (crc == questions_crc(header, (unsigned int)m, daemon->namebuff))
-		    m = process_reply(daemon, header, now, last_server, (unsigned int)m);
+		    m = process_reply(header, now, last_server, (unsigned int)m);
 		  
 		  break;
 		}
@@ -809,6 +836,8 @@ unsigned char *tcp_request(struct daemon *daemon, int confd, time_t now,
 	  if (m == 0)
 	    m = setup_reply(header, (unsigned int)size, addrp, flags, daemon->local_ttl);
 	}
+
+      check_log_writer(NULL);
       
       c1 = m>>8;
       c2 = m;
@@ -823,7 +852,7 @@ static struct frec *allocate_frec(time_t now)
 {
   struct frec *f;
   
-  if ((f = (struct frec *)malloc(sizeof(struct frec))))
+  if ((f = (struct frec *)whine_malloc(sizeof(struct frec))))
     {
       f->next = frec_list;
       f->time = now;
@@ -837,7 +866,7 @@ static struct frec *allocate_frec(time_t now)
 /* if wait==NULL return a free or older than TIMEOUT record.
    else return *wait zero if one available, or *wait is delay to
    when the oldest in-use record will expire. */
-struct frec *get_new_frec(struct daemon *daemon, time_t now, int *wait)
+struct frec *get_new_frec(time_t now, int *wait)
 {
   struct frec *f, *oldest;
   int count;
@@ -918,7 +947,7 @@ static struct frec *lookup_frec_by_sender(unsigned short id,
 }
 
 /* A server record is going away, remove references to it */
-void server_gone(struct daemon *daemon, struct server *server)
+void server_gone(struct server *server)
 {
   struct frec *f;
   
